from abc import ABCMeta, abstractmethod
from typing import Iterator

import regex as re
import phonenumbers
from .fact import Fact


class ParseResult:
    def __init__(self, value, span):
        self.value = value
        self.span = span


    def to_fact(self, fact_type, doc_path):
        return Fact(
            fact_type=fact_type,
            fact_value=self.value,
            doc_path=doc_path,
            spans=[self.span]
        )


class AbstractParser(metaclass=ABCMeta):

    @abstractmethod
    def parse(self) -> Iterator[ParseResult]:
        raise NotImplemented


class ContactPhoneParserHighRecall:
    #High Recall. Gets most of the phone numbers, output has a lot of noisy data.
    #a bit email-specific

    def __init__(self, input_text, months):
        self.text = input_text
        self.months = months


    def parse(self):
        """
        Finds all the phone numbers.
        Arguments:
        text -- A string of text to parse phone numbers from.
        Returns:
        numbers -- A list containing the numbers found.
        """
        months = self.months  # ?\:?[0-9]{2}( ?:[0-9]{2}){0,2} (GMT)?\+[0-9]{2} ?\:? ?[0-9]{2})
        pattern = re.compile(r'(?<![0-9]{2}[\./][0-9]{2}[\./])(?<![0-9]{2} [\./] [0-9]{2} [\./] )(?<!Date: )(?<!Date : )(?:\b|\+|\+\s)\d{1,4}(\s\(\s[0-9]+\s\)\s?)?[ -]?\(?\d{2,4}\)?(?:[ -]?\d{1,4}?\b)+(\s\([0-9]+\))?(?! ?:? ?[0-9]{2}( ?: ?[0-9]{2}){0,2} (GMT)?\+[0-9]{2} ?\:? ?[0-9]{2})')
        for numb in pattern.finditer(self.text):
            match_string = numb.group(0)
            if len(match_string) > 5:
                not_a_date = True
                for month in months:
                    if re.search("[0-9]{1,2} ?" + month + "\.? ?$", self.text[:numb.span()[0]].lower()) or re.search(month + " ?\.?\s[0-9]{1,2} ?\,? ?" + "$", self.text[:numb.span()[0]].lower()):
                        not_a_date = False
                        break
                if not_a_date:
                    replaced = match_string.replace(' ', '').replace('-', '').replace('+', '').replace(')', '').replace('(', '')
                    yield ParseResult(value=replaced, span=numb.span())

class ContactPhoneParserHighPrecision:
    #the output is mainly phone numbers. Excludes more complicated versions.
    def __init__(self, input_text):
        self.text = input_text

    def parse(self):
        """
        Finds all the phone numbers.
        Arguments:
        text -- A string of text to parse phone numbers from.
        Returns:
        numbers -- A list containing the numbers found.
        """
        pattern = re.compile(r'(?<![0-9a-zA-Z\:_&=\/#%@$&;])[+]*[-\s0-9]{0,4}[(]{0,1}[0-9]{1,4}[)]{0,1}[-\s0-9]*(?![0-9a-zA-Z\:_&=\/#%@$&;])')
        for numb in pattern.finditer(self.text):
            match_string = numb.group(0).strip()
            if len(re.sub("[^0-9]", "", match_string)) > 6 and not re.search("\n", match_string):
                replaced = match_string.replace(' ', '').replace('-', '').replace('+', '')
                yield ParseResult(value=replaced, span=numb.span()) 

class ContactPhoneParserStrict:
    """
    The output is phone numbers verified by phonenumbers library.
    country_code ==> list or str of the country codes, e.g. ["RU", "GB", "DE", "FI", "LV", "LT", "SE"] #Russia, Great Britannia, Germany, Finland, Latvia, Lithuania, Sweden]
    input_text ==> text from where the numbers are parsed out
    number ==> the number to be checked by library phonenumbers

    Notes:
    if country_code is wrong, it changes that (NB! in case the nr has suunakood, otherwise it's no good...)
    Also, with sentence "Maksekorraldusele märkida viitenumber 2800049900 ning selgitus .", the "2800049900" is a valid nr in "GB", but not with "EE" or "RU". So be careful when adding country_codes!

    """
    def __init__(self, input_text, country_code=["EE", "RU"]):
        self.text = input_text
        self.country_code = country_code
        
    def valid_and_possible(self, number):

        try:
            if type(self.country_code) == str:
                nr = phonenumbers.parse(number, self.country_code)
                if phonenumbers.is_valid_number(nr): #it’s in an assigned exchange and has the right number of digits
                    return True
            else: #country_code is a list 
                for code in self.country_code:
                    nr = phonenumbers.parse(number, code)
                    if phonenumbers.is_valid_number(nr):
                        return True
        except:
            return False
        return False

    def parse(self):
        """
        Finds all the phone numbers.
        Arguments:
        text -- A string of text to parse phone numbers from.
        Returns:
        numbers -- A list containing the numbers found.
        """
        pattern = re.compile(r'(?<![0-9a-zA-Z])[+]*[-\s0-9]{0,4}[(]{0,1}[\s0-9]{1,5}[)]{0,1}[-\s0-9]*(?![0-9a-zA-Z])')
        for numb in pattern.finditer(self.text):
            match_string = numb.group(0).strip()
            if len(re.sub("[^0-9]", "", match_string)) > 6 and not re.search("\n", match_string):
                if self.valid_and_possible(match_string):
                    yield ParseResult(value=match_string.replace(' ', '').replace('-', '').replace('+', ''), span=numb.span())  

class ContactEmailParser(AbstractParser):
    def __init__(self, input_text):
        self.text = input_text


    def parse(self):
        """
        Finds all the emails in the text.
        Arguments:
        text -- A string of text containing text to parse emails from.
        Returns:
        emails -- a list containing the emails found.
        """
        pattern = re.compile(r'([^\s]+\s?\.\s?)?[^\s@]+\s?@\s?[^@.\s]+\s?[.]\s?[^\s]{1,20}')
        for email in pattern.finditer(self.text):
            yield ParseResult(value=email.group(0).replace('mailto:', '').replace(' ', ''), span=email.span())


class ContactEmailNamePairParser(AbstractParser):
    def __init__(self, input_text):
        self.text = input_text


    def parse(self):
        """
        Finds all the emails in the text.
        Arguments:
        text -- A string of text containing text to parse emails from.
        Returns:
        emails -- a list containing the emails found.
        """
        pattern = re.compile(r'(?<=(\s|:|"|^))\p{L}+\s\p{L}+"?\s(< )?([^\s]+\s?\.\s?)?[^\s@]+\s?@\s?[^@.\s]+\s?[.]\s?[^\s]{1,20}')
        for email in pattern.finditer(self.text):
            email_splitted = email.group(0).split(" ")
            if len(email_splitted[1]) > 2 and len(email_splitted[0]) > 2 and email_splitted[0].title() == email_splitted[0] and email_splitted[1].title() == email_splitted[1]:
                yield ParseResult(value=email.group(0), span=email.span())


class AddressParser(AbstractParser):
    def __init__(self, text, stanza_entities, language):
        self.text = text
        self.stanza_entities = stanza_entities
        self.language = language


    @staticmethod
    def reescape(text):
        """
        https://stackoverflow.com/questions/43662474/reversing-pythons-re-escape 
        """
        return re.sub(r'\\(.)', r'\1', text)


    @property
    def _text_with_entity_types(self):
        """
        Joins entity words with their entity type/tag, e.g input 'Peeter läks Tallinnasse' outputs to 'Peeter_PER läks Tallinnasse_LOC'
        """
        replaced_text = self.text
        # sort stanza entities and check if they actually exist in text
        ents_to_replace = [(ent, len(ent.text)) for ent in self.stanza_entities if ent.text in self.text]
        for ent in sorted(ents_to_replace, key=lambda l: l[1]):
            ent = ent[0]
            tokens = ent.text.split(' ')
            modified_ent = '{}_{}'.format('_'.join(tokens), ent.type)
            # holy shit this is ugly, but it works...
            replaced_text = replaced_text.replace(" " + ent.text + " ", " " + modified_ent + " ")
        return replaced_text


    def _parse_ru(self):
        pattern = re.compile(
            r'(?:\b)([^\s]+_LOC\s?,?\s?((д\.|дом)\s?)?\s?[0-9]+\s?([а-яА-Я]|\s?/\s?[0-9]+|-[а-я])?)[\s\n,.]')
        for addr in pattern.finditer(self._text_with_entity_types):
            value = addr.group(1).replace('_LOC', '').replace('_', ' ')
            match = re.search(re.escape(value), self.text)
            if not match:
                try:
                    match = re.search(".".join(value.split()), self.text)
                    yield ParseResult(value=value.strip(), span=match.span())
                except:
                    continue
            else:
                yield ParseResult(value=value.strip(), span=match.span())


    def _parse_others(self):
        """
        dummy method for English, Estonian, Arabic
        TODO: proper regular expression for matching addresses in each language, except for Estonian
        """
        pattern = re.compile(r'\s([^\s]+_LOC\s?\s?[0-9]+)(_[A-Z]+)?')
        for addr in pattern.finditer(self._text_with_entity_types):
            value = addr.group(1).replace('_LOC', '').replace('_', ' ')
            match = re.search(re.escape(value), self.text)
            if not match:
                try:
                    match = re.search(".".join(value.split()), self.text)
                    yield ParseResult(value=value.strip(), span=match.span())
                except:
                    continue
            else:
                yield ParseResult(value=value.strip(), span=match.span())


    def parse(self):
        # pattern to search potential addresses
        if self.language == 'ru':
            return self._parse_ru()
        else:
            return self._parse_others()



import copy
import json
from typing import List

import numpy as np
import regex as re

from .exceptions import BoundedListEmpty


class Concatenator:
    """
    Class for concatenating BOUNDS across mails.
    """

    ENTITIES_TO_BEAUTIFY = ["ORG", "LOC", "GPE"]  # abbreviations and lemmas are checked when loading the data
    DEFAULT_MLP_OUTPUT_DIR = ""
    DEFAULT_ABBREVIATIONS_DIR = "data/abbreviations.json"


    def __init__(self, resource_dir=DEFAULT_MLP_OUTPUT_DIR, abbr_dir=DEFAULT_ABBREVIATIONS_DIR, logging_level="error"):
        self.resource_dir = resource_dir
        self.abbr_dir = abbr_dir
        self.abbreviations = dict()
        self.get_abbreviations()
        self.bounded = list()

        self.persona_infos = dict()  # {'Anna Puu': {'435432334': 'PHONE', '32793726354': 'PHONE', 'Новороссия': 'LOC', 'annapuu@gmail.com': 'EMAIL'} }
        self.no_personas_infos = list()
        self.unsure_infos = list()
        self.just_pers_infos = list()


    def get_abbreviations(self):
        abbrs = dict()
        with open(self.abbr_dir, 'r') as j:
            abbrs = json.loads(j.read())
        abbrs_ = dict()
        for key in abbrs:
            abbrs_[key.lower()] = abbrs[key].lower()
        self.abbreviations = abbrs_


    def beautify_entity_value(self, entity_value):
        """
        replace abbreviations with it's values
        """
        if entity_value.lower() in self.abbreviations:
            entity_value = self.abbreviations[entity_value.lower()].strip().title()
            return entity_value
        r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in entity_value.lower().split()]) + "|" + ".*".join([re.escape(s) for s in entity_value.lower().split()[::-1]]) + ")(\s|$)")
        parts_matched = list(filter(r.match, self.abbreviations.keys()))
        if parts_matched:
            entity_value = self.abbreviations[parts_matched[0]].strip().title()
            return entity_value
        entity_value_ = ""
        for word in entity_value.split():
            word = word.strip()
            if word.lower() in self.abbreviations:
                entity_value_ += " " + self.abbreviations[word.lower()].strip().title()
            else:
                entity_value_ += " " + word.title()
        return entity_value_.strip().strip("»«\*")


    def _abbreviations(self):
        return self.abbreviations


    def _bounded(self):
        return self.bounded


    def _persona_infos(self):
        return self.persona_infos


    def _no_personas_infos(self):
        return self.no_personas_infos


    def _unsure_infos(self):
        return self.unsure_infos


    def _just_pers_infos(self):
        return self.just_pers_infos


    def unique(self, bounded: list) -> list:
        """
        #GET UNIQUE DICTIONARIES, ERASE SUBSETS (WITH SAME KEYS)
        #first one BOUNDED is a subset of another one OR keys are the same {XXX, 'Новороссия': 'LOC'} and {XXX,'Новороссия': 'ORG'} (in this case entity_type is chosen randomly (whichever is first))
        """
        bounded = list(np.unique(np.array(bounded).astype(str)))
        bounds_to_delete_ids = []
        concatenated_bounds = []
        bounded_ = list()

        # Some BOUNDED with only one PER can be a subset of sth with two PERs, first take them out.
        for i, bound1 in enumerate(bounded):
            bound1 = json.loads(re.sub("\'", "\"", str(bound1)))
            per_values = [k for k, v in bound1.items() if v == "PER"]
            if len(per_values) == 1:
                if per_values[0] in self.persona_infos:
                    del bound1[per_values[0]]
                    self.persona_infos[per_values[0]].update(bound1)
                else:
                    del bound1[per_values[0]]
                    self.persona_infos[per_values[0]] = bound1
            else:
                bounded_.append(bound1)
        # Search for subsets
        for i, bound1 in enumerate(bounded_):
            for j, bound2 in enumerate(bounded_):
                if i != j and i not in bounds_to_delete_ids and j not in bounds_to_delete_ids:
                    if bound1.items() < bound2.items() or bound1.keys() <= bound2.keys():
                        bounds_to_delete_ids += [i]
                    elif bound1.items() > bound2.items() or bound1.keys() >= bound2.keys():
                        bounds_to_delete_ids += [j]
                    elif len(set([k for k, v in bound1.items() if v == "PER"])) != 0 and set([k for k, v in bound1.items() if v == "PER"]) == set([k for k, v in bound2.items() if v == "PER"]):
                        bound1.update(bound2)
                        bounds_to_delete_ids += [j]
            concatenated_bounds += [bound1]
        # return filtered BOUNDEDs
        filtered_bounds = []
        for i, bound1 in enumerate(concatenated_bounds):
            if i not in bounds_to_delete_ids:
                filtered_bounds += [bound1]
        return filtered_bounds


    @staticmethod
    def check_matching(parts_matched):
        all_match = True  # ["Tom", "Tom Marvolo", "Tom Marvolo Riddle"] == True, #["Tom", "Tom Marvolo", "Tom Kuusk"] == False
        for p_m1 in parts_matched:
            shortened1 = ""
            if len(p_m1.split()) == 1:
                shortened1 = "(^|\s)" + p_m1[0] + ".*"
            else:
                for k in p_m1.split()[:-1]:
                    shortened1 += k.strip(".") + ".*"
                shortened1 += p_m1.split()[-1]
            for p_m2 in parts_matched:
                shortened2 = ""
                if len(p_m2.split()) == 1:
                    shortened2 = "(^|\s)" + p_m2[0] + ".*"
                else:
                    for k in p_m2.split()[:-1]:
                        shortened2 += k.strip(".") + ".*"
                    shortened2 += p_m2.split()[-1]
                if len(p_m1.split()) == 3 and len(p_m2.split()) == 3:
                    if not (re.search(
                            "(^|\s)(" + ".*".join([re.escape(s) for s in p_m1.split()]) + "|" + ".*".join([re.escape(s) for s in p_m1.split()[::-1]]) + "|" + p_m1.split()[1] + ".*" + p_m1.split()[2] + ".*" + p_m1.split()[0] + "|" + p_m1.split()[0] + ".*" + p_m1.split()[2] + ".*" + p_m1.split()[
                                1] + ")(\s|$)", p_m2) or re.search(
                        "(^|\s)(" + ".*".join([re.escape(s) for s in p_m2.split()]) + "|" + ".*".join([re.escape(s) for s in p_m2.split()[::-1]]) + "|" + p_m2.split()[1] + ".*" + p_m2.split()[2] + ".*" + p_m2.split()[0] + "|" + p_m2.split()[0] + ".*" + p_m2.split()[2] + ".*" + p_m2.split()[
                            1] + ")(\s|$)", p_m1) or re.search(shortened1, shortened2) or re.search(shortened2, shortened1)):
                        return False
                else:
                    if not (re.search("(^|\s)(" + ".*".join([re.escape(s) for s in p_m1.split()]) + "|" + ".*".join([re.escape(s) for s in p_m1.split()[::-1]]) + ")(\s|$)", p_m2) or re.search(
                            "(^|\s)(" + ".*".join([re.escape(s) for s in p_m2.split()]) + "|" + ".*".join([re.escape(s) for s in p_m2.split()[::-1]]) + ")(\s|$)", p_m1) or re.search(shortened1, shortened2) or re.search(shortened2, shortened1)):
                        return False
        return all_match


    def unique_keys(self, pers_dict):
        """
        Concatenate key-value pairs that has similar keys.
        Eg. "Tom Riddle" + "Tom Marvolo Riddle"
        Used in function divide
        """
        keys = copy.deepcopy(list(pers_dict.keys()))
        for key in keys:
            keys_ = copy.deepcopy(list(pers_dict.keys()))
            if key in keys_:
                keys_.remove(key)
                splitted_key = key.split()
                if len(splitted_key) > 1:
                    shortened = ""
                    if "." in key:
                        for name in splitted_key[:-1]:
                            shortened += name[0].strip() + ".*"
                        shortened += splitted_key[-1]
                    else:
                        for name in splitted_key[:-1]:
                            shortened += name[0].strip() + ". "
                        shortened += splitted_key[-1]
                    if len(splitted_key) == 3:
                        if shortened:
                            r = re.compile(
                                "(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" + splitted_key[
                                    1] + "|" + shortened + ")(\s|$)")
                    else:
                        if shortened:
                            r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + shortened + ")(\s|$)")
                else:
                    r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")(\s|$)")
                parts_matched = list(filter(r.search, keys_))  # if PER is "Tom Riddle", and there´s already "Tom Marvolo Riddle"
                if len(parts_matched) == 1:
                    if len(parts_matched[0]) >= len(key):
                        pers_dict[parts_matched[0]].update(pers_dict[key])
                        del pers_dict[key]
                    else:
                        pers_dict[key].update(pers_dict[parts_matched[0]])
                        del pers_dict[parts_matched[0]]
                elif len(parts_matched) > 1:
                    all_match = self.check_matching(parts_matched)
                    if all_match:
                        longest = max(parts_matched, key=len)
                        parts_matched.remove(longest)
                        pers_dict[longest].update(pers_dict[key])
                        del pers_dict[key]
                        for p_m in parts_matched:
                            if p_m != longest:
                                pers_dict[longest].update(pers_dict[p_m])
                                del pers_dict[p_m]
                    else:
                        if len(splitted_key) > 1:
                            shortened = ""
                            if "." in key:
                                for name in splitted_key[:-1]:
                                    shortened += name[0].strip() + ".*"
                                shortened += splitted_key[-1]
                            if len(splitted_key) == 3:
                                if shortened:
                                    r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" +
                                                   splitted_key[1] + "|" + shortened + ")$")
                                else:
                                    r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" +
                                                   splitted_key[1] + ")$")
                            else:
                                if shortened:
                                    r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + shortened + ")$")
                                else:
                                    r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")$")
                        else:
                            r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")$")
                        parts_matched = list(filter(r.search, parts_matched))
                        if len(parts_matched) == 1:
                            if len(parts_matched[0]) >= len(key):
                                pers_dict[parts_matched[0]].update(pers_dict[key])
                                del pers_dict[key]
                            else:
                                pers_dict[key].update(pers_dict[parts_matched[0]])
                                del pers_dict[parts_matched[0]]
                        elif len(parts_matched) > 1:
                            all_match = self.check_matching(parts_matched)
                            if all_match:
                                longest = max(parts_matched, key=len)
                                parts_matched.remove(longest)
                                pers_dict[longest].update(pers_dict[key])
                                del pers_dict[key]
                                for p_m in parts_matched:
                                    if p_m != longest:
                                        pers_dict[longest].update(pers_dict[p_m])
                                        del pers_dict[p_m]
        return pers_dict


    def matching_key(self, per_value):
        """
        If there`s no equal key, check, if there`s partly equal keys (per_value is a subset) and those are subsets or superssets of each other.

        Example:
        If persona_infos.keys() = ["Tom", "Tom Marvolo", "Tom Marvolo Riddle", "Tom Kuusk", "Муравьев Леонид Иванович"]
        then:
        per_value = "Tom Marvolo" ==> parts_matched = ["Tom Marvolo", "Tom Marvolo Riddle"] ==> all_match, return the list,
        per_value = "Tom" ==> parts_matched = ["Tom", "Tom Marvolo", "Tom Kuusk"] ==> all_do_not_match, return None,
        per_value = "Леонид Муравьев" ==> partc_matched = ["Муравьев Леонид Иванович"] == all_match, return the list
        """
        splitted_key = per_value.split()
        if len(splitted_key) > 1:
            shortened = ""
            if "." in per_value:
                for name in splitted_key[:-1]:
                    shortened += name[0].strip() + ".*"
                shortened += splitted_key[-1]
            if len(splitted_key) == 3:
                if shortened:
                    r = re.compile(
                        "(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" + splitted_key[
                            1] + "|" + shortened + ")(\s|$)")
                else:
                    r = re.compile(
                        "(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" + splitted_key[
                            1] + ")(\s|$)")
            else:
                if shortened:
                    r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + shortened + ")(\s|$)")
                else:
                    r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")(\s|$)")
        else:
            r = re.compile("(^|\s)(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")(\s|$)")

        parts_matched = list(filter(r.search, self.persona_infos.keys()))  # if PER is "Tom Riddle", and there´s already "Tom Marvolo Riddle"
        if len(parts_matched) == 1:
            return parts_matched
        elif len(parts_matched) > 1:
            all_match = self.check_matching(parts_matched)
            if all_match:
                return parts_matched
            else:
                if len(splitted_key) > 1:
                    shortened = ""
                    if "." in per_value:
                        for name in splitted_key[:-1]:
                            shortened += name[0].strip() + ".*"
                        shortened += splitted_key[-1]
                    if len(splitted_key) == 3:
                        if shortened:
                            r = re.compile(
                                "^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" + splitted_key[
                                    1] + "|" + shortened + ")$")
                        else:
                            r = re.compile(
                                "^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + splitted_key[1] + ".*" + splitted_key[2] + ".*" + splitted_key[0] + "|" + splitted_key[0] + ".*" + splitted_key[2] + ".*" + splitted_key[
                                    1] + ")$")
                    else:
                        if shortened:
                            r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + "|" + shortened + ")$")
                        else:
                            r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")$")
                else:
                    r = re.compile("^(" + ".*".join([re.escape(s) for s in splitted_key]) + "|" + ".*".join([re.escape(s) for s in splitted_key[::-1]]) + ")$")
                parts_matched = list(filter(r.search, parts_matched))
                all_match = self.check_matching(parts_matched)
                if all_match:
                    return parts_matched
        return None


    def divide(self):
        """
        #divide bounded to persona_infos, no_personas_infos, just_pers_infos and unsure_infos
        """

        for f_bound_ in self.bounded:
            f_bound = copy.deepcopy(f_bound_)
            if len(f_bound.items()) > 1:
                per_values = [k for k, v in f_bound.items() if v == "PER"]
                if len(per_values) == len(f_bound.items()):
                    self.just_pers_infos.append(f_bound)
                elif len(per_values) == 0:
                    self.no_personas_infos.append(f_bound)
                elif len(per_values) == 1:  # Should not exist anymore, but just checking.
                    if per_values[0] in self.persona_infos:
                        del f_bound[per_values[0]]
                        self.persona_infos[per_values[0]].update(f_bound)
                    else:
                        del f_bound[per_values[0]]
                        self.persona_infos[per_values[0]] = f_bound
                else:
                    self.unsure_infos.append(f_bound)


    def entity_per_check(self, new_unsure_infos, entity_pers_dict):
        """
        Check whether some entity occurs only with some certain PER.
        Example:
        entity_pers_dict = {'Gurkhul':[["Terry Pratchett", "Joe Abercrombie"], ["Joe Abercrombie", "J.K. Rowling"]], #NB! We can see that J.A. is always with Gurkhul.
                            'Loss':[["Marten Kuningas", "Martena Kuninganna"], ["Prince", "Queen"]]} #NB! The entity_type doesn´t matter here anymore.
        new_unsure_infos = [{'Joe Abercrombie': 'PER', 'Gurkhul': 'LOC', "J.K. Rowling": 'PER'},
                        {'Joe Abercrombie': 'PER', 'Gurkhul': 'LOC', 'Terry Pratchett': 'PER'},
                        {"Marten Kuningas": 'PER', 'Loss': 'LOC',  "Martena Kuninganna": 'PER'},
                        {"Prince": 'PER', 'Loss': 'LOC', "Queen": 'PER'}]
        persona_infos = {'Aleksander Great': {'76883266': 'PHONE',  'aleksandersuur356eKr@mail.ee': 'EMAIL', 'Vana-Makedoonia': 'LOC'},
                        'Terry Pratchett': {'Discworld': 'LOC', '12345678': 'PHONE', 'Ankh-Morpork': 'LOC'}}

        ==>

        just_pers_infos =  [{'Aleksander Great': 'PER', 'Terry Pratchett': 'PER'},
                            {'Joe Abercrombie': 'PER', 'Terry Pratchett': 'PER'},
                            {'Joe Abercrombie': 'PER', "J.K. Rowling": 'PER'}]
        unsure_infos = [{"Marten Kuningas": 'PER', 'Loss': 'LOC',  "Martena Kuninganna": 'PER'},
                        {"Prince": 'PER', 'Loss': 'LOC', "Queen": 'PER'}]
        persona_infos = {'Aleksander Great': {'76883266': 'PHONE',  'aleksandersuur356eKr@mail.ee': 'EMAIL', 'Vana-Makedoonia': 'LOC'},
                        'Terry Pratchett': {'Discworld': 'LOC', '12345678': 'PHONE', 'Ankh-Morpork': 'LOC'},
                        'Joe Abercrombie': {'Gurkhul': 'LOC'}}

        save unsure_infos.
        """
        actually_has_a_per = dict()  # {"Discworld" : "Terry Pratchett"}
        for entity_value, per_values in entity_pers_dict.items():
            always_with_the_entity_pers = list(set.intersection(*per_values))
            if len(always_with_the_entity_pers) == 1:
                actually_has_a_per[entity_value] = always_with_the_entity_pers[0]

        double_checked_unsure_infos = list()
        for u_bound in new_unsure_infos:
            u_bound_ = dict()
            for entity_value, entity_type in u_bound.items():
                if entity_value not in actually_has_a_per:
                    u_bound_[entity_value] = entity_type
                else:
                    if actually_has_a_per[entity_value] in self.persona_infos:
                        if entity_value not in self.persona_infos[actually_has_a_per[entity_value]]:
                            self.persona_infos[actually_has_a_per[entity_value]].update({entity_value: entity_type})
                    else:
                        self.persona_infos[actually_has_a_per[entity_value]] = {entity_value: entity_type}
            per_values = [k for k, v in u_bound_.items() if v == "PER"]
            if len(per_values) == len(u_bound_.items()):
                self.just_pers_infos.append(u_bound_)
            else:
                double_checked_unsure_infos.append(u_bound_)

        self.unsure_infos = double_checked_unsure_infos
        del new_unsure_infos
        del double_checked_unsure_infos


    def check_unsure_infos(self):
        """
        #check if some entites in BOUNDEDs in unsure_infos are same in persona_infos under the BOUNDED PERs. Delete those as connection found already. Remaining is still unsure (to which PER does the LOC/EMAIL etc belong),
        if no entity is left, add to self.just_pers_infos.

        Example:
        unsure_infos = [{'Aleksander Great': 'PER',  'aleksandersuur356eKr@mail.ee': 'EMAIL', 'Vana-Makedoonia': 'LOC', 'Terry Pratchett': 'PER', 'Discworld': 'LOC', '12345678': 'PHONE', 'Ankh-Morpork': 'LOC'},
        {'Joe Abercrombie': 'PER', 'Gurkhul': 'LOC', 'Terry Pratchett': 'PER', 'Ankh-Morpork': 'LOC'}]
        persona_infos = {'Aleksander Great': {'76883266': 'PHONE',  'aleksandersuur356eKr@mail.ee': 'EMAIL', 'Vana-Makedoonia': 'LOC'},
                        'Terry Pratchett': {'Discworld': 'LOC', '12345678': 'PHONE', 'Ankh-Morpork': 'LOC'}}

        ==>

        just_pers_infos =  [{'Aleksander Great': 'PER', 'Terry Pratchett': 'PER'}]
        new_unsure_infos = [{'Joe Abercrombie': 'PER', 'Gurkhul': 'LOC', 'Terry Pratchett': 'PER'}]
        persona_infos = {'Aleksander Great': {'76883266': 'PHONE',  'aleksandersuur356eKr@mail.ee': 'EMAIL', 'Vana-Makedoonia': 'LOC'},
                        'Terry Pratchett': {'Discworld': 'LOC', '12345678': 'PHONE', 'Ankh-Morpork': 'LOC'}}
        """
        new_unsure_infos = list()
        entity_pers_dict = dict()  # {"Discworld" : [["Terry Pratchett", "Joe Abercrombie"], ["Terry Pratchett", "J.K. Rowling"]]}
        for i, u_bounded_ in enumerate(self.unsure_infos):
            u_bounded = copy.deepcopy(u_bounded_)
            per_values = [k for k, v in u_bounded.items() if v == "PER"]  # PERs in that BOUNDED fact.
            to_delete = list()
            values_to_replace = []
            for per_value in per_values:
                intersections = list()
                if per_value in self.persona_infos:
                    intersection = set(u_bounded) & set(self.persona_infos[per_value])
                    if intersection != set():
                        intersections.append(intersection)
                else:
                    parts_matched = self.matching_key(per_value)
                    if parts_matched:
                        # print(per_value,  max(parts_matched+[per_value], key=len), parts_matched+[per_value])
                        values_to_replace.append((per_value, max(parts_matched + [per_value], key=len)))
                        intersection = set(u_bounded)
                        for part_matched in parts_matched:
                            intersection = intersection & set(self.persona_infos[part_matched])
                        if intersection != set():
                            intersections.append(intersection)
                if intersections != []:
                    to_delete.append(set.intersection(*intersections))

            for delete_set in to_delete:
                for delete in delete_set:
                    if delete in u_bounded and delete not in per_values:
                        del u_bounded[delete]

            for old_per, new_per in values_to_replace:
                del u_bounded[old_per]
                u_bounded[new_per] = "PER"
            if len(per_values) == len(u_bounded.items()):
                self.just_pers_infos.append(u_bounded)
            else:
                per_values = [k for k, v in u_bounded.items() if v == "PER"]  # PERs in that renewed BOUNDED fact.
                for entity_value, entity_type in u_bounded.items():
                    if entity_type != "PER":
                        if entity_value in entity_pers_dict:
                            entity_pers_dict[entity_value] += [set(per_values)]
                        else:
                            entity_pers_dict[entity_value] = [set(per_values)]

                new_unsure_infos.append(u_bounded)

        return new_unsure_infos, entity_pers_dict


    def concatenate(self):
        if not self.bounded:
            raise BoundedListEmpty("The class has no BOUNDED facts loaded yet.")
        self.bounded = self.unique(self.bounded)
        self.divide()
        self.persona_infos = self.unique_keys(self.persona_infos)
        new_unsure_infos, entity_pers_dict = self.check_unsure_infos()
        self.entity_per_check(new_unsure_infos, entity_pers_dict)
        self.persona_infos = self.unique_keys(self.persona_infos)


    def saveable_list(self, dictionary_name, type_name):
        new_list = list()
        for infos_dict in dictionary_name:
            new_dict = {"type": type_name}
            for entity_value in infos_dict:
                if infos_dict[entity_value] in new_dict:
                    new_dict[infos_dict[entity_value]] += [entity_value]
                else:
                    new_dict[infos_dict[entity_value]] = [entity_value]
            new_list.append(new_dict)
        return new_list


    def to_json(self):
        container = []

        for person in self._persona_infos():
            new_dict = {"type": "person_info", "PER": person}
            for key in self._persona_infos()[person]:
                if self._persona_infos()[person][key] in new_dict and self._persona_infos()[person][key] != "PER":
                    new_dict[self._persona_infos()[person][key]] += [key]
                else:
                    new_dict[self._persona_infos()[person][key]] = [key]
            container.append(new_dict)

        no_per = self.saveable_list(self._no_personas_infos(), "no_per_close_entities")
        if no_per: container.append(no_per)

        whose_entities = self.saveable_list(self._unsure_infos(), "unsure_whose_entities")
        if whose_entities: container.append(whose_entities)

        close_pers = self.saveable_list(self._just_pers_infos(), "close_persons")
        if close_pers: container.append(close_pers)

        return container


    def from_json(self, mlp_outputs: List[dict]):
        for output in mlp_outputs:
            for fact in output["texta_facts"]:
                if fact["fact"] == "BOUNDED":
                    bound = {}
                    bounded_facts = json.loads(re.sub("\'", "\"", fact['str_val']))
                    lemmas = json.loads(re.sub("\'", "\"", fact["lemma"]))
                    for entity_type in bounded_facts:
                        if entity_type in self.ENTITIES_TO_BEAUTIFY:
                            for i, entity_value in enumerate(lemmas[entity_type]):
                                if entity_value != "UNK":
                                    bound[self.beautify_entity_value(entity_value)] = entity_type
                                else:
                                    bound[self.beautify_entity_value(bounded_facts[entity_type][i])] = entity_type
                        else:
                            if entity_type != "EMAIL":
                                for entity_value in bounded_facts[entity_type]:
                                    bound[entity_value.strip().strip("»«\*").title()] = entity_type
                            else:
                                for entity_value in bounded_facts[entity_type]:
                                    bound[entity_value.strip("»«\*").strip()] = entity_type
                    self.bounded.append(bound)



import json
from typing import List


class Fact:

    def __init__(self, fact_type: str, fact_value: str, doc_path: str = None, fact_lemma: str = None, spans: list = [[0, 0]]):
        self.fact_type = fact_type
        self.fact_value = fact_value
        self.spans = spans
        self.doc_path = doc_path
        self.fact_lemma = fact_lemma


    @staticmethod
    def from_json(list_of_facts: List[dict]):
        container = []
        for fact_json in list_of_facts:
            fact_type = fact_json.get("fact")
            spans = json.loads(fact_json.get("spans"))
            value = fact_json.get("str_val")
            field_path = fact_json.get("doc_path")
            fact = Fact(fact_type=fact_type, fact_value=value, doc_path=field_path, spans=spans)
            container.append(fact)
        return container


    def to_json(self):
        container = {
            "str_val": self.fact_value,
            "spans": json.dumps(self.spans),
            "fact": self.fact_type,
            "doc_path": self.doc_path,
        }
        if self.fact_lemma:
            container["lemma"] = self.fact_lemma
        return container



# -*- coding: utf8 -*-
import json
import regex as re


class EntityMapper():

    def __init__(self, entity_data):
        self.entities, self.longest_entity = self._load_entities(entity_data)
        self.entity_map = self._create_entity_map()


    def _create_entity_map(self):
        '''
        Takes the entities and creates an entity_map

        Returns:
        entity_map -- A dictionary with entities as keys and type as the value of the key
        '''
        entity_map = {}
        for entity_type, entity_list in self.entities.items():
            for entity in entity_list:
                if entity not in entity_map:
                    entity_map[entity] = entity_type
        return entity_map


    def _load_entities(self, entity_files):
        '''
        Loads entities from a file into a dict and stores the lenght of the longest entity.

        Arguments:
        entity_data -- a list that contains the type and location of the entities, for example: [('addr', 'addresses.txt')]

        Returns:
        entities -- A dict with the entity type and its contents  

        longest_entity -- the longest entity from all the entities
        '''

        entities = {}
        longest_entity = 0

        for entity_file in entity_files:
            with open(entity_file, encoding='utf8') as fh:
                try:
                    # Try to read as json
                    for k, v in json.loads(fh.read()).items():
                        v = [self._normalize_string(w) for w in v]
                        if k not in entities:
                            entities[k] = {}

                        for w in v:
                            entities[k][w] = True

                        entity_len = max([len(w.split(' ')) for w in v])
                        if entity_len > longest_entity:
                            longest_entity = entity_len

                except:
                    # If fails, read as txt file
                    entities[entity_file] = {}
                    for v in fh.readlines():
                        v = self._normalize_string(v)
                        entities[entity_file][v] = True

                        entity_len = len(v.split(' '))
                        if entity_len > longest_entity:
                            longest_entity = entity_len

        return entities, longest_entity


    @staticmethod
    def _normalize_string(entity):
        '''
        Takes in an argument string 'entity' and returns it after lowering it and stripping whitespace
        '''
        entity = entity.lower().strip()
        return entity


    def map_entities(self, text, doc_path='', entity_types=None):
        '''
        Maps entities in given text.

        Arguments:
        text -- A string of text to map. For example: u'I live on Street 1, not Street 2'

        doc_path='' -- Path to the given document where the text was taken from.        


        Returns:
        filtered_mappings -- A sorted (descending by value lenght) dict that has redundant matches removed.
        For example: {'addr': [{'span': (57, 69), 'doc_path': '', 'value': u'Street 2'}, {'span': (41, 51), 'doc_path': '', 'value': u'Street 1'}]}
        '''
        if not entity_types:
            entity_types = list(self.entities.keys())

        mappings = []

        text = self._normalize_string(text)

        for n in range(1, self.longest_entity + 1):
            ngrams = self._ngrams(text, n)
            for ngram in ngrams:
                ngram = ' '.join(ngram)
                # This is kinda hacky, but it kinda fixes Stanza tokenization errors.
                ngram = ngram.replace(".", "").strip()
                # Checks if ngram is in the entity map. If so, define its type and add it to the mappings
                if ngram in self.entity_map:
                    entity_type = self.entity_map[ngram]
                    if entity_type in entity_types:
                        mapping = {'entity_type': entity_type, 'entity': ngram}
                        if mapping not in mappings:
                            mappings.append(mapping)
        mappings_dict = {}
        for mapping in mappings:
            if mapping['entity_type'] not in mappings_dict:
                mappings_dict[mapping['entity_type']] = []
            mappings_dict[mapping['entity_type']].append({'value': mapping['entity']})

        mappings = self._add_spans(mappings_dict, text, doc_path)
        mappings = self._remove_redundant_matches(mappings)

        return mappings


    @staticmethod
    def _remove_redundant_matches(mappings):
        '''
        Takes in the mappings of entities and removes redundant matches.

        Arguments: 
        mappings -- A dict that contains the facts about the entities, including the span, doc_path, type and value.
        For example: {'addr': [{'span': (41, 51), 'doc_path': '', 'value': u'Street 1'}, {'span': (57, 69), 'doc_path': '', 'value': u'Street 2'}]}

        Returns:
        filtered_mappings -- A sorted (descending by value lenght) dict that has redundant matches removed.
        For example: {'addr': [{'span': (57, 69), 'doc_path': '', 'value': u'Street 2'}, {'span': (41, 51), 'doc_path': '', 'value': u'Street 1'}]}
        '''
        filtered_mappings = {}

        for entity_type in mappings.keys():
            filtered_matches = []
            # Sorts by the lenght of the values in the entity type in descending order.
            sorted_matches = sorted(
                mappings[entity_type], key=lambda k: len(k['value']), reverse=True)

            for match in sorted_matches:
                m_start = match['span'][0]
                m_end = match['span'][1]
                redundant = False

                for filtered_match in filtered_matches:
                    fm_start = filtered_match['span'][0]
                    fm_end = filtered_match['span'][1]

                    if m_start >= fm_start and m_end <= fm_end:
                        redundant = True
                        break

                if redundant == False:
                    filtered_matches.append(match)

            filtered_mappings[entity_type] = filtered_matches

        return filtered_mappings


    @staticmethod
    def _add_spans(mappings, text, doc_path):
        '''
        Adds spans (character number locations of entities) and the document path to the entity dict.

        Arguments:
        mappings -- A dict that contains entity types, and the values of different entities under that type.
        For example: {'addr': [{'value': u'Street 1'}, {'value': u'Street 2'}]}

        text -- The input text given for mapping

        doc_path -- Path to the doc containing the entity

        Returns:
        facts -- A dict that contains the facts about the entities, including the span, doc_path, type and value.
        For example: {'addr': [{'span': (41, 51), 'doc_path': '', 'value': u'Street 1'}, {'span': (57, 69), 'doc_path': '', 'value': u'Street 2'}]}
        '''
        facts = {}

        for entity_type in mappings.keys():
            values_with_spans = []
            for entity in mappings[entity_type]:
                # get entity value, like u'Street 1' and compile a regex pattern
                pattern = re.compile(entity['value'])
                for match in pattern.finditer(text):
                    value_with_span = {'value': entity['value']}
                    value_with_span['span'] = (match.start(), match.end())
                    if doc_path:
                        value_with_span['doc_path'] = doc_path
                    values_with_spans.append(value_with_span)

            if entity_type not in facts:
                facts[entity_type] = []
            facts[entity_type] = values_with_spans
        return facts


    @staticmethod
    def _ngrams(inp, n):
        '''
        Takes in an input text and returns ngrams

        Arguments: 
        inp -- Input text

        n -- The range of iteration

        Returns:
        returns a list that contains words as tuples, the word count depends on the argument 'n',
        for example if n is 1, the output will contain only one word per list value, ex: ([(u'example',), (u'string',), (u'here',)])
        but if n is 2, the output will contain two words per list value, ex: ([(u'example string',), (u'string here',))])
        '''
        if type(inp) in [bytes, str]:
            inp = inp.split(' ')
        return zip(*[inp[i:] for i in range(n)])


if __name__ == '__main__':
    entity_data = {'addr': 'data/addresses.txt'}
    em = EntityMapper(entity_data)
    print(em.map_entities(
        u'Selles stringis on aadressid, milleks on Ülikooli 2 ning Kentmanni 13'))



import json
import math
from typing import List, Optional

import regex as re
from lang_trans.arabic import buckwalter
from pelecanus import PelicanJson

from .entity_mapper import EntityMapper
from .fact import Fact
from .parsers import AddressParser, ContactEmailNamePairParser, ContactEmailParser, ContactPhoneParserHighPrecision, \
    ContactPhoneParserHighRecall, ContactPhoneParserStrict
from .russian_transliterator import Transliterate


russian_transliterator = Transliterate()


class Document:
    """
    Class for isolating the different values of MLP like pos_tags, lemmas etc
    and their formatting. For adding/changing the values, create a function with the name
    of the analyzer like "forms()" that handles populating the __forms attribute and a "get_forms()"
    function to format it. In the end add "get_forms()" to the to_json() function.
    """
    langs_to_transliterate = ["ru", "ar"]

    FACT_NAME_EMAIL = "EMAIL"
    FACT_NAME_ADDRESS = "ADDR"
    FACT_NAME_PHONE_HIGH_RECALL = "PHONE_high_recall"
    FACT_NAME_PHONE_HIGH_PRECISION = "PHONE_high_precision"
    FACT_NAME_PHONE_STRICT = "PHONE_strict"
    FACT_NAMES_NER = ("PER", "ORG", "GPE", "LOC")
    FACT_NAME_NAMEMAIL = "NAMEMAIL"
    FACT_NAME_BOUNDED = "BOUNDED"

    KNOWN_ENTITIES = (FACT_NAME_EMAIL, FACT_NAME_ADDRESS, FACT_NAME_PHONE_STRICT, FACT_NAME_NAMEMAIL) + FACT_NAMES_NER

    CLOSE_FACT_DISTANCE = 150


    def __init__(
            self,
            original_text: str,
            dominant_language_code: str,
            analysis_lang: str,
            stanza_sentences: [list],
            stanza_entities,
            concat_resources: dict,
            entity_mapper: Optional[EntityMapper] = None,
            doc_path: str = "text",
            json_doc: dict = None,
            analyzers: list = [],
            error: str = "",
    ):

        self.original_text = original_text
        self.doc_path = doc_path
        self.analyzers = analyzers
        self.dominant_language_code = dominant_language_code
        self.analysis_lang = analysis_lang
        self.error = error
        self.json_doc = json_doc

        self.entity_mapper = entity_mapper
        self.stanza_sentences = stanza_sentences
        self.stanza_words = [word for sentence in self.stanza_sentences for word in sentence]
        self.stanza_entities = stanza_entities

        self.concat_resources = concat_resources

        self.__words = []
        self.__lemmas = []
        self.__pos_tags = []
        self.__transliteration = []
        self.__texta_facts: List[Fact] = []

        self.__handle_existing_facts()
        self.words()


    def __get_doc_path(self, field: str) -> str:
        """
        :param field: Whether the doc_path uses the text or lemmas field.
        :return: MLP representation of the doc_path
        """
        content = f"{self.doc_path}_mlp.{field}"
        return content


    def __handle_existing_facts(self):
        """
        Add existing texta_facts inside the document into the private
        fact container variable so that they wouldn't be overwritten.
        """
        if self.json_doc and "texta_facts" in self.json_doc:
            existing_facts = self.json_doc.get("texta_facts", [])
            facts = Fact.from_json(existing_facts)
            for fact in facts:
                self.add_fact(fact)


    @staticmethod
    def handle_null_values_in_facts(facts: List[dict]):
        container = []
        if facts:
            for fact in facts:
                new_fact = {key: value for key, value in fact.items() if value is not None}
                container.append(new_fact)
        return container


    @staticmethod
    def remove_duplicate_facts(facts: List[dict]):
        if facts:
            facts = Document.handle_null_values_in_facts(facts)
            set_of_jsons = {json.dumps(fact, sort_keys=True, ensure_ascii=False) for fact in facts}
            without_duplicates = [json.loads(unique_fact) for unique_fact in set_of_jsons]
            return without_duplicates
        else:
            return []


    def facts_to_json(self) -> dict:
        facts = [fact.to_json() for fact in self.__texta_facts]
        unique_facts = Document.remove_duplicate_facts(facts)
        return {"texta_facts": unique_facts}


    def add_fact(self, fact: Fact):
        self.__texta_facts.append(fact)


    @staticmethod
    def edit_doc(doc: dict, doc_path: str, new_value) -> dict:
        """
        :param doc: Original dictionary into which you add content.
        :param doc_path: Dot separated field paths to the nested value you wish to add.
        :param new_value: Value you wish to add into the nested path.
        :return: Dictionary with the new value added into it according to the path.
        """
        wrapper = PelicanJson(doc)
        path = doc_path.split(".")
        wrapper.set_nested_value(path, new_value, force=True)
        return wrapper.convert()


    def document_to_json(self, use_default_doc_path=True) -> dict:
        """
        :param use_default_doc_path: Normal string values will be given the default path for facts but for dictionary input you already have them.
        """
        list_of_path_keys = self.doc_path.split(".")
        root_key = "{}_mlp".format(list_of_path_keys[-1])
        path_to_mlp = list_of_path_keys[:-1] + [root_key] if len(list_of_path_keys) > 1 else [root_key]
        mlp_result = self.to_json(use_default_doc_path)
        nested_dict_wrapper = PelicanJson(self.json_doc)
        nested_dict_wrapper.set_nested_value(path_to_mlp, mlp_result["text"], force=True)
        nested_dict_wrapper.set_nested_value(["texta_facts"], mlp_result["texta_facts"], force=True)

        return nested_dict_wrapper.convert()


    def to_json(self, use_default_doc_path=True) -> dict:
        container = dict()
        container["text"] = self.get_words(ssplit="sentences" in self.analyzers)
        texta_facts = self.facts_to_json()
        container["language"] = {
            "detected": self.dominant_language_code,
            "analysis": self.analysis_lang
        }
        if self.error:
            container["error"] = self.error

        if "lemmas" in self.analyzers:
            container["lemmas"] = self.get_lemma()
        if "pos_tags" in self.analyzers:
            container["pos_tags"] = self.get_pos_tags()
        # if "sentiment" in self.analyzers: container["sentiment"] = self.get_sentiment()
        if "transliteration" in self.analyzers and self.__transliteration:
            container["transliteration"] = self.get_transliteration()
        if use_default_doc_path:
            for fact in texta_facts["texta_facts"]:
                fact["doc_path"] = "text.text"
        return {"text": container, **texta_facts}


    def lemmas(self):
        for sent in self.stanza_sentences:
            self.__lemmas.append([word.lemma.replace("_", "") if word and word.lemma else "X" for word in sent])


    def get_lemma(self) -> str:
        sentences = []
        for sent_lemmas in self.__lemmas:
            sentences.append(" ".join([a.strip() for a in sent_lemmas]))

        if "sentences" in self.analyzers:
            return "\n".join(sentences)
        else:
            return " ".join(sentences)


    def words(self):
        for sent in self.stanza_sentences:
            self.__words.append([word.text for word in sent])


    def sentences(self):
        pass


    def get_words(self, ssplit=False) -> str:
        if ssplit:
            return "\n".join([" ".join(sent_words) for sent_words in self.__words])
        else:
            return " ".join([" ".join(sent_words) for sent_words in self.__words])


    def pos_tags(self):
        self.__pos_tags = [word.xpos if word and word.xpos and word.xpos != "_" else "X" if word.xpos == "_" else "X"
                           for word in self.stanza_words]


    def get_pos_tags(self) -> str:
        return " ".join([a.strip() for a in self.__pos_tags])


    def entities(self):
        """
        Retrieves list-based entities.
        """
        text = self.get_words()
        lemmas = self.get_lemma()

        hits = self.entity_mapper.map_entities(text)
        lemma_hits = self.entity_mapper.map_entities(lemmas, entity_types=["CURRENCY"])

        # make facts
        for entity_type, entity_values in hits.items():
            for entity_value in entity_values:
                new_fact = Fact(
                    fact_type=entity_type,
                    fact_value=entity_value["value"],
                    doc_path=self.__get_doc_path("text"),
                    spans=[[entity_value["span"][0], entity_value["span"][1]]]
                )
                self.__texta_facts.append(new_fact)

        for entity_type, entity_values in lemma_hits.items():
            for entity_value in entity_values:
                new_fact = Fact(
                    fact_type=entity_type,
                    fact_value=entity_value["value"],
                    doc_path=self.__get_doc_path("lemmas"),
                    spans=[[entity_value["span"][0], entity_value["span"][1]]]
                )
                self.__texta_facts.append(new_fact)

        # declare the entities processed
        self.entities_processed = True


    def currency_sum(self):
        """
        Extracts currency + sum and sum + currency patterns from text using regexp.
        Saves extractions as facts.
        """
        text = self.get_words()
        currency_facts = [fact for fact in self.__texta_facts if fact.fact_type == "CURRENCY"]
        for fact in currency_facts:
            regexes = (
                f"{fact.fact_value} [0-9,\.]+",
                f"[0-9,\.]+ {fact.fact_value}[a-z]*"
            )
            for currency_regex in regexes:
                pattern = re.compile(currency_regex)
                for match in pattern.finditer(text):
                    fact_value = match.string[match.start():match.end()]
                    # recheck that string contains a number
                    if any(map(str.isdigit, fact_value)):
                        new_fact = Fact(
                            fact_type="CURRENCY_SUM",
                            fact_value=fact_value,
                            doc_path=self.__get_doc_path("text"),
                            spans=[match.start(), match.end()]
                        )
                        self.__texta_facts.append(new_fact)


    def emails(self):
        text = self.get_words()
        emails = ContactEmailParser(text).parse()
        self.__texta_facts.extend((email.to_fact(Document.FACT_NAME_EMAIL, self.__get_doc_path("text")) for email in emails))


    def phone_strict(self):
        text = self.get_words()
        phone_numbers_strict = ContactPhoneParserStrict(text).parse()
        self.__texta_facts.extend(
            (number.to_fact(Document.FACT_NAME_PHONE_STRICT, self.__get_doc_path("text")) for number in phone_numbers_strict))


    def phone_high_recall(self):
        text = self.get_words()
        phone_numbers = ContactPhoneParserHighRecall(text, months=self.concat_resources["months"]).parse()
        self.__texta_facts.extend(
            (number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.__get_doc_path("text")) for number in phone_numbers))


    def phone_high_precision(self):
        text = self.get_words()
        phone_numbers_high_precision = ContactPhoneParserHighPrecision(text).parse()
        self.__texta_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_PRECISION, self.__get_doc_path("text")) for number in
                                   phone_numbers_high_precision))


    def addresses(self):
        text = self.get_words()
        addresses = AddressParser(text, self.stanza_entities, self.dominant_language_code).parse()
        self.__texta_facts.extend((addr.to_fact(Document.FACT_NAME_ADDRESS, self.__get_doc_path("text")) for addr in addresses))


    def transliteration(self):
        if self.dominant_language_code in Document.langs_to_transliterate:
            for word in self.stanza_words:
                if self.dominant_language_code == "ru":
                    translit_word = self._transliterate_russian_word(word)
                elif self.dominant_language_code == "ar":
                    translit_word = self._transliterate_arabic_word(word)
                self.__transliteration.append(translit_word)


    @staticmethod
    def _transliterate_russian_word(word):
        translit_word = russian_transliterator([word.text.strip()])
        try:
            translit_word = translit_word[0].strip()
        except IndexError:
            translit_word = word.text.strip()
        return translit_word


    @staticmethod
    def _transliterate_arabic_word(word):
        translit_word = buckwalter.transliterate(word.text.strip())
        if not translit_word:
            translit_word = word.text.strip()
        return translit_word


    def get_transliteration(self) -> str:
        return " ".join(['X' if not a.strip() else a for a in self.__transliteration])


    def entity_lemmas(self, entity_value):
        lemmas = ""
        splitted = entity_value.split(" ")
        for i, word in enumerate(self.stanza_words):
            if word.text == splitted[0]:
                if len(splitted) > 1:
                    isthatphrase = True
                    j = i
                    for entity_word in splitted[1:]:
                        j += 1
                        if j < len(self.stanza_words) and entity_word != self.stanza_words[j].text:
                            isthatphrase = False
                        if j >= len(self.stanza_words):
                            isthatphrase = False
                    if isthatphrase:
                        lemmas += word.lemma
                        for i_, entity_word in enumerate(splitted[1:]):
                            lemmas += " " + self.stanza_words[i_ + i + 1].text
                        return lemmas
                else:
                    return word.lemma
        return lemmas


    def ner(self):
        tokenized_text = self.get_words()
        known_entities = Document.FACT_NAMES_NER
        not_entities = self.concat_resources["not_entities"]
        for entity in self.stanza_entities:
            if entity.text.lower() in not_entities:
                continue
            if entity.type in known_entities:
                # finds the closest spans in tokenized text
                # this is because stanza returns spans from non-tokenized text
                pattern = re.compile(re.escape(entity.text))  # Use re.escape to avoid trouble with special characters existing in the text.
                matching_tokenized_spans = [(match.start(), match.end()) for match in pattern.finditer(tokenized_text)]
                best_matching_span = None
                best_matching_distance = math.inf
                non_tokenized_span = (entity.start_char, entity.end_char)
                # matching spans are always equal or larger
                for span in matching_tokenized_spans:
                    span_distance = (span[0] - non_tokenized_span[0]) + (span[1] - non_tokenized_span[1])
                    if abs(span_distance) < best_matching_distance:
                        best_matching_distance = abs(span_distance)
                        best_matching_span = span
                # create and append fact
                # ignore facts whose match fails
                if best_matching_span:
                    new_fact = Fact(
                        fact_type=entity.type,
                        fact_value=entity.text,
                        doc_path=self.__get_doc_path("text"),
                        spans=[best_matching_span]
                    )
                    self.__texta_facts.append(new_fact)


    def namemail(self):
        """
        Find name-email pairs.

        """
        text = self.get_words()
        email_name_pairs = ContactEmailNamePairParser(text).parse()  # bounded -> str "name mail"
        self.__texta_facts.extend(
            (emailpair.to_fact(Document.FACT_NAME_NAMEMAIL, self.__get_doc_path("text")) for emailpair in email_name_pairs))


    def remove_duplicate_facts_by_span(self, facts):
        """if there are emailpairs, then:
        [{fact_type: "NAMEMAIL", value: "Aleksander Great aleksandersuur 356eKr@mail.ee", spans(30,60)}, {fact_type: "PER", value: "Aleksander Great", spans:(30,40)}] ==>
        [{fact_type: "MAIL", value: "aleksandersuur 356eKr@mail.ee", spans(40,60)}, {fact_type: "PER", value: "Aleskander Great", spans:(30,40)}]
        NAMEMAIL is used, because sometimes the ner_tagger only gets the PER or the EMAIL. This makes double sure that we get most of the entities and there is no overlaping entities
        (as it was before).
        """
        starts = []
        ends = []
        new_facts = []
        facts_values = list()
        for fact in facts:
            fact.fact_value = fact.fact_value.strip("><\)\(:;-\.,\!\?")
        for fact in facts:
            facts_values += [fact.fact_value]
            if fact.fact_type == Document.FACT_NAME_NAMEMAIL:
                splitted_value = fact.fact_value.split(" ")
                name_fact_value = " ".join(splitted_value[:2])
                name_fact = Fact(
                    fact_type="PER",
                    fact_value=name_fact_value.strip("><\)\(:;-\.,\!\?"),
                    doc_path=self.__get_doc_path("text"),
                    spans=[(fact.spans[0][0], fact.spans[0][0] + len(name_fact_value))]
                )

                email_fact_value = " ".join(splitted_value[2:])
                email_fact = Fact(
                    fact_type="EMAIL",
                    fact_value=email_fact_value.strip("><\)\(:;-\.,\!\?"),
                    doc_path=self.__get_doc_path("text"),
                    spans=[(fact.spans[0][1] - len(email_fact_value), fact.spans[0][1])]
                )

                new_facts += [email_fact, name_fact]
                starts += [email_fact.spans[0][0], name_fact.spans[0][0]]
                ends += [email_fact.spans[0][1], name_fact.spans[0][1]]

        for fact in facts:
            if fact.fact_type != Document.FACT_NAME_NAMEMAIL and fact.spans[0][0] not in starts and fact.spans[0][
                1] not in ends:
                fact.fact_value = fact.fact_value.strip("><\)\(:;-\.,\!\?")
                new_facts += [fact]
        new_facts_values = list()
        for fact in new_facts:
            new_facts_values += [fact.fact_value]
        return new_facts


    @staticmethod
    def clean_similar_in_strval(bound1):
        """
        #'PER': 'Павел Губарев', 'Павел Юрьевич Губарев' ==> 'PER': Павел Юрьевич Губарев
        #{'Народного Совета по промышленности и торговле': 'ORG', 'Народного Совета': 'ORG'} ==> {'Народного Совета по промышленности и торговле': 'ORG'}
        """
        bound = {}
        for entity_type in bound1:
            for entity_value in bound1[entity_type]:
                bound[entity_value] = entity_type
        new_bound = {}
        no_is = []
        for i, item1 in enumerate(bound.items()):
            issubitem = False
            for j, item2 in enumerate(bound.items()):
                if i != j and i not in no_is and j not in no_is:
                    if re.search("(^|\s)" + re.escape(item1[0].strip("><)(:;-.,!?")) + "(\s|$|\@[a-z]{4,5}\.[a-z]{2})",
                                 item2[0].strip("><)(:;-.,!?")) and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(
                        item2[0].strip("><\)\(:;-\.,\!\?")):
                        if item2[1] in new_bound:
                            new_bound[item2[1]] += [item2[0].strip("><)(:;-.,!?")]
                        else:
                            new_bound[item2[1]] = [item2[0].strip("><)(:;-.,!?")]
                        no_is += [i]
                        issubitem = True
                    else:  # each word of this entity exists also in the other one.
                        notsameparts = False
                        for part_of_entity in item1[0].split(" "):
                            if not re.search("(^|\s)" + re.escape(part_of_entity.strip("><)(:;-.,!?")) + "(\s|$)",
                                             item2[0].strip("><)(:;-.,!?")):
                                notsameparts = True
                                break
                        if not notsameparts and len(item1[0].strip("><)(:;-.,!?")) <= len(
                                item2[0].strip("><)(:;-.,!?")):
                            if item2[1] in new_bound:
                                new_bound[item2[1]] += [re.sub("(\'|\")", "*", item2[0].strip("><)(:;-.,!?"))]
                            else:
                                new_bound[item2[1]] = [re.sub("(\'|\")", "*", item2[0].strip("><)(:;-.,!?"))]
                            no_is += [i]
                            issubitem = True
            if not issubitem:
                if item1[1] in new_bound:
                    new_bound[item1[1]] += [re.sub("(\'|\")", "*", item1[0].strip("><\)\(:;-\.,\!\?"))]
                else:
                    new_bound[item1[1]] = [re.sub("(\'|\")", "*", item1[0].strip("><\)\(:;-\.,\!\?"))]
        for entity_type in new_bound:
            new_bound[entity_type] = list(set(new_bound[entity_type]))
        return new_bound


    def space_between_ok(self, text, first_spans, second_spans):
        """
        tests if between the entities there is no beginning of a new letter
        """
        if first_spans[1] > 15:  # s@yandex.ru > : Hello, Steven! | l@gmail.com :\nHello, Stephanie!
            text_between = text[first_spans[1] - 15:second_spans[0]]
            pat = re.compile("\p{L}@\p{L}{3,10}\.\p{L}{2,3} {0,2}>? ?:\s")
            if pat.search(text_between):
                return False
        text_between = text[first_spans[1]:second_spans[0]]
        if re.search(self.concat_resources["space_between_not_ok"], text_between):
            return False
        for month in self.concat_resources["months"]:
            if re.search("[0-9]{1,2} ?" + month + " ?\.? ?[0-9]{4}( г\.)? ?, ?в? [0-9]{2} ?: ?[0-9]{2}",
                         text_between.lower()):
                return False
        return True


    @staticmethod
    def key_value_single_pairs(d1):
        """d1 = {'a': [1,2], 'b':[3,4]} --> [('a', 1), ('a',2), ('b',3), ('b',4)]"""
        d1_pairs = []
        for key, values in d1.items():
            for value in values:
                d1_pairs += [(key, value)]
        return d1_pairs


    def bound_close_ones(self, facts):
        """
        Bound facts found in that email, that has distance of  <=self.CLOSE_FACT_DISTANCE and that has self.space_between_ok True, together.
        """
        facts = sorted(facts, key=lambda item: item.spans[0][0])
        new_bound = {'doc_path': self.__get_doc_path("text"), 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {},
                     "str_values": []}
        new_bounded = []
        for fact in facts:
            if len(new_bound["spans"]) == 0:
                new_bound["spans"] += fact.spans
                new_bound["str_val"][fact.fact_type] = [fact.fact_value]
                new_bound["str_values"] = [(fact.fact_value, fact.fact_type)]
            else:
                if abs(fact.spans[0][0] - new_bound["spans"][-1][
                    -1]) < self.CLOSE_FACT_DISTANCE and self.space_between_ok(self.get_words(), new_bound["spans"][-1],
                                                                              fact.spans[0]):
                    new_bound["spans"] += fact.spans
                    new_bound["str_values"] += [(fact.fact_value, fact.fact_type)]
                    if fact.fact_type in new_bound["str_val"]:
                        new_bound["str_val"][fact.fact_type] += [fact.fact_value]
                    else:
                        new_bound["str_val"][fact.fact_type] = [fact.fact_value]
                else:  # start a new bound
                    if len(new_bound["spans"]) > 1:
                        new_bounded += [new_bound]
                    new_bound = {'doc_path': self.__get_doc_path("text"), 'fact': 'BOUNDED', 'spans': [], "str_val": {},
                                 "str_values": []}
                    new_bound["spans"] += fact.spans
                    new_bound["str_values"] = [(fact.fact_value, fact.fact_type)]
                    new_bound["str_val"][fact.fact_type] = [fact.fact_value]
        if len(new_bound["spans"]) > 1:
            new_bounded += [new_bound]
        return new_bounded


    def remove_overlaping_in_bounded(self, new_bounded):
        """
        if bounded has two entities with spans starting on the same index, pick the longest one.
        For example:
        'spans': '[[174, 194], [174, 185]]',
   'str_val': "{'EMAIL': ['89179012978@mail.ru]'], 'PHONE': ['89179012978']}"},
       clean_similar_in_strval: strvals contain each other, concatenate.
       """
        new_bounded_coinciding_erased = []
        for new_bound in new_bounded:
            to_delete_is = []
            for i, span1 in enumerate(new_bound["spans"]):
                for j, span2 in enumerate(new_bound["spans"]):
                    if span1[0] == span2[0] and i != j:
                        if span1[1] > span2[1]:
                            to_delete_is += [j]
                        elif span1[1] < span2[1]:
                            to_delete_is += [i]
            new_bound_erased_coinciding = {'doc_path': self.__get_doc_path("text"), 'fact': Document.FACT_NAME_BOUNDED, 'spans': [],
                                           'str_val': {}}
            for i, span1 in enumerate(new_bound["spans"]):
                if i not in to_delete_is:
                    new_bound_erased_coinciding["spans"] += [span1]

                    if new_bound["str_values"][i][1] in new_bound_erased_coinciding["str_val"]:
                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] += [
                            new_bound["str_values"][i][0]]
                    else:
                        new_bound_erased_coinciding["str_val"][new_bound["str_values"][i][1]] = [
                            new_bound["str_values"][i][0]]
            new_bound_erased_coinciding["str_val"] = self.clean_similar_in_strval(
                new_bound_erased_coinciding["str_val"])
            hasmanyentities = False
            for entity_type in new_bound_erased_coinciding["str_val"]:
                if len(new_bound_erased_coinciding["str_val"][entity_type]) > 1:
                    hasmanyentities = True
                    break
            if len(new_bound_erased_coinciding["spans"]) > 1 and (
                    hasmanyentities or len(new_bound_erased_coinciding["str_val"].items()) > 1):
                new_bounded_coinciding_erased += [new_bound_erased_coinciding]
        return new_bounded_coinciding_erased


    def concatenate_subset_bounds(self, new_bounded):
        """
        When one BOUNDED is a subset of another one by strings, concatenate
        As this is the last finishing touch, add lemmas to the string_value
        """
        concatenated_bound_is = []
        new_bounds = list()
        for i, new_bound in enumerate(new_bounded):
            issubset = False
            for j, another_bound in enumerate(new_bounded):
                set_comparison = set(self.key_value_single_pairs(new_bound["str_val"])) <= set(self.key_value_single_pairs(another_bound["str_val"]))
                if j not in concatenated_bound_is and i not in concatenated_bound_is and i != j and set_comparison:
                    issubset = True
                    another_bound["spans"] += new_bound["spans"]
                    concatenated_bound_is += [i]
            if not issubset:
                lemmas = {}
                for key in new_bound["str_val"]:
                    for entity in new_bound["str_val"][key]:
                        lemma = self.entity_lemmas(entity)
                        if not lemma:
                            lemma = "UNK"
                        if key in lemmas:
                            lemmas[key] += [lemma]
                        else:
                            lemmas[key] = [lemma]
                new_fact = Fact(
                    fact_type=new_bound["fact"],
                    fact_value=str(new_bound["str_val"]),  ##add lemmas
                    fact_lemma=str(lemmas),
                    doc_path=self.__get_doc_path("text"),
                    spans=new_bound["spans"]
                )
                new_bounds.append(new_fact)
        return new_bounds


    def remove_old_bounded_and_add_phone_high_recall(self, facts):
        """
        remove previous BOUNDED facts
        remove previous phone facts
        add phoneparser made for emails (ContactPhoneParserHighRecall)
        """
        new_facts = [fact for fact in facts if not re.match("PHONE|" + Document.FACT_NAME_BOUNDED, fact.fact_type)]
        text = self.get_words()
        phone_numbers = ContactPhoneParserHighRecall(text, months=self.concat_resources["months"]).parse()
        new_facts.extend(
            (number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.__get_doc_path("text")) for number in phone_numbers))
        return new_facts


    def bounded(self):
        facts = self.remove_duplicate_facts_by_span(self.__texta_facts)
        facts = self.remove_old_bounded_and_add_phone_high_recall(
            facts)  # bounded uses only ContactPhoneParserHighRecall results and doesn't look at old BOUNDED facts.
        new_bounded = self.bound_close_ones(facts)
        new_bounded = self.remove_overlaping_in_bounded(new_bounded)
        new_bounded = self.concatenate_subset_bounds(new_bounded)
        for new_bound in new_bounded:
            self.__texta_facts.append(new_bound)
